from __future__ import division
import numpy as np


#TODO: document me
def regression(X, T, regression_type, **kwargs):
    r"""
    """
    if hasattr(regression_type, '__call__'):
        regression_closure = regression_type(X, T, **kwargs)

        # note that, self is required as this closure is assigned to an object
        def regression_object_method(x):
            return regression_closure(x)

        return regression_closure
    else:
        raise ValueError("regression_type can only be: a closure defining "
                         "a particular regression technique. Several examples"
                         " of such closures can be found in "
                         "`menpo.sdm.functions` (mlr, pcr, pls, ccr, ...).")


#TODO: document me
def mlr(X, T):
    r"""
    Multivariate Linear Regression
    """
    XX = np.dot(X.T, X)
    XX = (XX + XX.T) / 2
    XT = np.dot(X.T, T)
    R = np.linalg.solve(XX, XT)

    def mlr_fitting(x):
        return np.dot(x, R)

    return mlr_fitting


#TODO: document me
def mlr_svd(X, T, variance=None):
    r"""
    Multivariate Linear Regression using SVD decomposition
    """
    R, _, _, _ = _svd_regression(X, T, variance=variance)

    def mlr_svd_fitting(x):
        return np.dot(x, R)

    return mlr_svd_fitting


#TODO: document me
def mlr_pca(X, T, variance=None):
    r"""
    Multivariate Linear Regression using PCA reconstructions
    """
    R, _, _, V = _svd_regression(X, T, variance=variance)

    def mlr_svd_fitting(x):
        x = np.dot(np.dot(x, V.T), V)
        return np.dot(x, R)

    return mlr_svd_fitting


#TODO: document me
def mlr_pca_weights(X, T, variance=None):
    r"""
    Multivariate Linear Regression using PCA weights
    """
    R, _, _, V = _svd_regression(X, T, variance=variance)
    W = np.dot(X, V.T)
    R, _, _, _ = _svd_regression(W, T)

    def mlr_svd_fitting(x):
        w = np.dot(x, V.T)
        return np.dot(w, R)

    return mlr_svd_fitting


def _svd_regression(X, T, variance=None):
    if variance is not None and not (0 < variance <= 1):
        raise ValueError("variance must be set to a number between 0 and 1.")

    U, s, V = np.linalg.svd(X)
    if variance:
        total = sum(s)
        acc = 0
        for j, y in enumerate(s):
            acc += y
            if acc / total >= variance:
                r = j+1
                break
    else:
        tol = np.max(X.shape) * np.spacing(np.max(s))
        r = np.sum(s > tol)
    U = U[:, :r]
    s = 1 / s[:r]
    V = V[:r, :]
    R = np.dot(np.dot(V.T * s, U.T), T)

    return R, U, s, V



from __future__ import division
import abc
import os.path as path
import numpy as np
from menpo.image import ShapeImage
from menpo.io.base import (Importer, find_alternative_files,
                           map_filepath_to_importer)
import re


class SpatialImageImporter(Importer):
    r"""
    Base class for importing depth images. Depth images are defined by file
    types whereby the data lies explicitly on a grid. This grid pattern means
    the data can be interpreted as an image.

    Parameters
    ----------
    filepath : string
        An absolute filepath
    """

    def __init__(self, filepath):
        super(SpatialImageImporter, self).__init__(filepath)
        self.attempted_texture_search = False
        self.relative_texture_path = None
        self.trilist = None
        self.mask = None
        self.tcoords = None
        self.shape_image = None

    def _build_texture_and_landmark_importers(self):
        r"""
        Search for a texture and landmark file in the same directory as the
        mesh. If they exist, create importers for them.
        """
        if self.texture_path is None or not path.exists(self.texture_path):
            self.texture_importer = None
        else:
            # This import is here to avoid circular dependencies
            from menpo.io.extensions import image_types
            self.texture_importer = map_filepath_to_importer(self.texture_path,
                                                 image_types)

    def _search_for_texture(self):
        r"""
        Tries to find a texture with the same name as the depth image.

        Returns
        --------
        relative_texture_path : string
            The relative path to the texture or ``None`` if one can't be found
        """
        # Stop searching every single time we access the property
        self.attempted_texture_search = True
        # This import is here to avoid circular dependencies
        from menpo.io.extensions import image_types
        try:
            return find_alternative_files('texture', self.filepath,
                                          image_types)
        except ImportError:
            return None

    @property
    def texture_path(self):
        """
        Get the absolute path to the texture. Returns None if one can't be
        found. Makes it's best effort to find an appropriate texture by
        searching for textures with the same name as the mesh. Will only
        search for the path the first time ``texture_path`` is invoked.

        Sets the ``self.relative_texture_path`` attribute.

        Returns
        -------
        texture_path : string
            Absolute filepath to the texture
        """
        # Try find a texture path if we can
        if self.relative_texture_path is None and \
                not self.attempted_texture_search:
            self.relative_texture_path = self._search_for_texture()

        try:
            return path.join(self.folder, self.relative_texture_path)
        except AttributeError:
            return None

    @abc.abstractmethod
    def _build_image_and_mesh(self):
        r"""
        Abstract method that handles actually building an image. This involves
        reading the image from disk and doing any necessary processing.

        Should set the ``self.image`` attribute.
        """
        pass

    @abc.abstractmethod
    def _process_landmarks(self, original_image, landmark_group):
        r"""
        Abstract method that allows the landmarks to be transformed between
        the original image space and the depth image space. The space may
        be different because a texture mapped image is unlikely to be equal
        in size to the mesh.

        Parameters
        ----------
        original_image : :class:`menpo.image.base.Image`
            The original image that the landmarks belong to
        lmark_group : :class:`menpo.landmark.base.LandmarkGroup`
            The landmark group to transform

        Returns
        -------
        transformed_group : :class:`menpo.landmark.base.LandmarkGroup`
            The transformed landmark group.
        """
        pass

    def build(self):
        r"""
        Overrides the :meth:`build <menpo.io.base.Importer.build>` method.

        Builds the image and landmarks. Assigns the landmark set to the image.

        Returns
        -------
        image : :class:`menpo.image.base.Image`
            The image object with landmarks attached if they exist
        """
        # Build the image as defined by the overridden method and then search
        # for valid landmarks that may have been defined by the importer
        self._build_image_and_mesh()
        self._build_texture_and_landmark_importers()

        if self.texture_importer is not None:
            texture = self.texture_importer.build()
        else:
            texture = None

        self.image = ShapeImage(self.shape_image, mask=self.mask,
                                trilist=self.trilist, tcoords=self.tcoords,
                                texture=texture)
        return self.image


class BNTImporter(SpatialImageImporter):
    r"""
    Allows importing the BNT file format from the bosphorus dataset.
    This reads in the 5 channels (3D coordinates and texture coordinates),
    splits them appropriately and then triangulates the ``x`` and ``y``
    coordinates to create a surface. The texture path is also given in the file
    format.

    The file format specifies a 'bad_value' which is used to denote inaccurate
    data points. This value is replaced with ``np.nan``. The mesh is
    created with only those values that are not NaN. The depth image contains
    all values.

    Parameters
    ----------
    filepath : string
        Absolute filepath of the mesh.
    """

    def __init__(self, filepath):
        # Setup class before super class call
        super(BNTImporter, self).__init__(filepath)

    def _build_image_and_mesh(self):
        r"""
        Read the file in and parse appropriately. Includes reading the texture
        path.
        """
        with open(self.filepath, 'rb') as f:
            # Currently these are unused, but they are in the format
            # Could possibly store as metadata?
            n_rows = np.fromfile(f, dtype=np.uint16, count=1)
            n_cols = np.fromfile(f, dtype=np.uint16, count=1)
            bad_value = np.fromfile(f, dtype=np.float64, count=1)

            # Get integers and convert to valid string
            image_path_len = np.fromfile(f, dtype=np.uint16, count=1)
            texture_path = np.fromfile(f, dtype=np.uint8, count=image_path_len)
            texture_path = ''.join(map(chr, texture_path))

            # Get data and reshape (reshape in an odd order due to Matlab's
            # Fortran ordering). First three columns are 3D coordinates
            # and last two are 2D texture coordinates
            coords_len = np.fromfile(f, dtype=np.uint32, count=1)
            data = np.fromfile(f, dtype=np.float64, count=coords_len * 5.0)
            data = data.reshape([5, coords_len / 5.0]).T

        # Get the 3D coordinates
        shape_pixels = data[:, :3]
        # We want to remove the bad values because otherwise the mesh is not
        # renderable. We do this by replacing the bad value values with nan
        shape_pixels[shape_pixels == bad_value] = np.nan

        # The image contains all coordinates
        # Must be flipped LR due to Fortran ordering from Matlab
        # Must by flipped upside down due to image vs mesh ordering
        self.shape_image = np.fliplr(np.reshape(shape_pixels[:, :3][::-1],
                                                [n_rows, n_cols, 3]))
        self.mask = ~np.any(np.isnan(self.shape_image), axis=-1)

        # Use only those coordinates with do not contains nans
        valid_points = ~np.isnan(shape_pixels).any(axis=1)

        # Apparently the texture coordinates are upside down?
        #self.tcoords = data[:, -2:][valid_points]
        #self.tcoords[:, 1] = 1.0 - self.tcoords[:, 1]
        # struggling to interpret these - let the ShapeImage build them
        # instead.
        self.tcoords = None

        self.relative_texture_path = texture_path


class FIMImporter(SpatialImageImporter):
    r"""
    Allows importing floating point images as depth images.
    This reads in the shape in to 3 channels and then triangulates the
    ``x`` and ``y`` coordinates to create a surface. An example of this
    datatype is the aligned BU4D dataset.

    Parameters
    ----------
    filepath : string
        Absolute filepath of the mesh.
    """

    def __init__(self, filepath):
        # Setup class before super class call
        super(FIMImporter, self).__init__(filepath)

    def _build_image_and_mesh(self):
        r"""
        Read the file and parse it as necessary. Since the data lies on a grid
        we can triangulate the 2D coordinates to get a valid triangulation.

        The format does not specify texture coordinates.
        """
        with open(self.filepath, 'rb') as f:
            size = np.fromfile(f, dtype=np.uint32, count=3)
            data = np.fromfile(f, dtype=np.float32, count=np.product(size))
            data = data.reshape([size[0], size[1], size[2]])

        # Replace the zero buffer values with nan so that the image renders
        # nicely
        data[data == 0] = np.nan

        self.shape_image = data
        self.mask = ~np.isnan(self.shape_image)


class ABSImporter(SpatialImageImporter):
    r"""
    Allows importing the ABS file format from the FRGC dataset.

    The z-min value is stripped from the mesh to make it renderable.

    Parameters
    ----------
    filepath : string
        Absolute filepath of the mesh.
    """

    def __init__(self, filepath):
        # Setup class before super class call
        super(ABSImporter, self).__init__(filepath)

    def _build_image_and_mesh(self):
        r"""
        Read in the file and remove the z-min. Triangulate the 2D gridded
        coordinates to create a valid triangulation.
        """
        with open(self.filepath, 'r') as f:
            # Currently these are unused, but they are in the format
            # Could possibly store as metadata?
            # Assume first result for regexes
            re_rows = re.compile(u'([0-9]+) rows')
            n_rows = int(re_rows.findall(f.readline())[0])
            re_cols = re.compile(u'([0-9]+) columns')
            n_cols = int(re_cols.findall(f.readline())[0])


        # This also loads the mask
        #   >>> image_data[:, 0]
        image_data = np.loadtxt(self.filepath, skiprows=3, unpack=True)

        # Replace the lowest value with nan so that we can render properly
        image_data[image_data == np.min(image_data)] = np.nan

        self.shape_image = np.reshape(image_data[:, 1:], [n_rows, n_cols, 3])
        self.mask = np.reshape(image_data[:, 0], [n_rows, n_cols])



import abc
from copy import deepcopy
import os.path


class Vectorizable(object):
    """
    Interface that provides methods for 'flattening' an object into a
    vector, and restoring from the same vectorized form. Useful for
    statistical analysis of objects, which commonly requires the data
    to be provided as a single vector.
    """

    __metaclass__ = abc.ABCMeta

    @property
    def n_parameters(self):
        r"""
        The length of the vector that this Vectorizable object produces.

        type: int
        """
        return (self.as_vector()).shape[0]

    @abc.abstractmethod
    def as_vector(self):
        """
        Returns a flattened representation of the object as a single
        vector.

        Returns
        -------
        vector : (N,) ndarray
            The core representation of the object, flattened into a
            single vector.
        """
        pass

    @abc.abstractmethod
    def from_vector_inplace(self, vector):
        """
        Update the state of this object from it's vectorized state

        Parameters
        ----------
        vector : (N,) ndarray
            Flattened representation of this object.
        """
        pass

    def from_vector(self, vector):
        """
        Build a new instance of the object from it's vectorized state.


        ``self`` is used to fill out the missing state required to
        rebuild a full object from it's standardized flattened state. This
        is the default implementation, which is which is a
        ``deepcopy`` of the object followed by a call to
        :meth:`from_vector_inplace()`. This method can be overridden for a
        performance benefit if desired.

        Parameters
        ----------
        vector : (N,) ndarray
            Flattened representation of the object.

        Returns
        -------
        object : :class:`Vectorizable`
            An instance of the class.
        """
        self_copy = deepcopy(self)
        self_copy.from_vector_inplace(vector)
        return self_copy


class Targetable(object):
    r"""
    Interface for objects that can produce a 'target' PointCloud - which
    could for instance be the result of an alignment or a generation of a
    PointCloud instance from a shape model.

    Implementations must define sensible behavior for:

     - what a target is: target property
     - how to set a target: _target_setter
     - how to update the object after a target is set: _sync_state_from_target
     - how to produce a new target after the changes: _new_target_from_state

    Note that _sync_target_from_state() needs to be triggered as appropriate by
    subclasses e.g. when from_vector_inplace is called. This will in turn
    trigger _new_target_from_state(), which each subclass must implement.
    """
    __metaclass__ = abc.ABCMeta

    @property
    def n_dims(self):
        return self.target.n_dims

    @property
    def n_points(self):
        return self.target.n_points


    @abc.abstractproperty
    def target(self):
        pass

    def set_target(self, value):
        r"""
        Updates this alignment transform to point to a new target.
        """
        self._target_setter_with_verification(value)  # trigger the update
        self._sync_state_from_target()  # and a sync

    def _target_setter_with_verification(self, value):
        r"""
        Updates the target, checking it is sensible, without triggering a sync.

        Should be called by _sync_target_from_state with the new target value.
        """
        self._verify_target(value)
        self._target_setter(value)

    def _verify_target(self, new_target):
        # If the target is None (i.e. on construction) then dodge the
        # verification
        if self.target is None:
            return
        if new_target.n_dims != self.target.n_dims:
            raise ValueError(
                "The current target is {}D, the new target is {}D - new "
                "target has to have the same dimensionality as the "
                "old".format(self.target.n_dims, new_target.n_dims))
        elif new_target.n_points != self.target.n_points:
            raise ValueError(
                "The current target has {} points, the new target has {} "
                "- new target has to have the same number of points as the"
                " old".format(self.target.n_points, new_target.n_points))

    @abc.abstractmethod
    def _target_setter(self, new_target):
        r"""
        Sets the target to the new value. Does no synchronization. Note that
        it is advisable that _target_setter_with_verification is called from
        subclasses instead of this.
        """
        pass

    def _sync_target_from_state(self):
        new_target = self._new_target_from_state()
        self._target_setter_with_verification(new_target)

    @abc.abstractmethod
    def _new_target_from_state(self):
        r"""
        Returns a new target that is correct after changes to the object.
        """
        pass

    @abc.abstractmethod
    def _sync_state_from_target(self):
        r"""
        Synchronizes the object state to be correct after changes to the target.

        Called automatically from the target setter. This is called after the
        target is updated - only handle synchronization here.
        """
        pass


def menpo_src_dir_path():
    return os.path.split(os.path.abspath(__file__))[0]



import abc
import numpy as np
from scipy.spatial.distance import cdist


class BasisFunction(object):
    r"""
    An abstract base class for Basis functions. In the case, radial basis
    functions. They provide two methods, :meth:`apply`, which calculates the
    basis itself, and :meth:`jacobian_points`, which calculates the derivative
    of the basis wrt the coordinate system.

    Parameters
    ----------
    c : (L, D) ndarray
        The set of centers that make the basis. Usually represents a set of
        source landmarks.
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, c):
        self.c = c

    @abc.abstractmethod
    def apply(self, x):
        r"""
        Calculate the basis function on the given residuals. The input is the
        set of points the basis should be calculated for. The euclidean
        distance between ``x`` and the centers, ``c``, will be used as the
        residual.

        .. note::

            Divisions by zero are avoided and any zero residuals remain zero.

        Parameters
        ----------
        x : (N, D) ndarray
            Set of points to apply the basis to.

        Returns
        -------
        u : (N, L) ndarray
            The basis function applied to each distance,
            :math:`\lVert x - c \rVert`.
        """
        pass

    @abc.abstractmethod
    def jacobian_points(self, x):
        r"""
        Calculate the derivative of the basis function wrt the
        coordinate system.

        .. note::

            Divisions by zero are avoided and any zero residuals remain zero.

        Parameters
        ----------
        x : (N, D) ndarray
            Set of points to apply the basis to.

        Returns
        -------
        dudx : (N, L, D) ndarray
            Tensor representing the first order partial derivative
            of each points with respect to the centers, over each dimension.
        """
        pass


class R2LogR2(BasisFunction):
    r"""
    The :math:`r^2 \log{r^2}` basis function.

    The derivative of this function is :math:`2 r (\log{r^2} + 1)`.

    .. note::

        :math:`r = \lVert x - c \rVert`

    Parameters
    ----------
    c : (L, D) ndarray
        The set of centers that make the basis. Usually represents a set of
        source landmarks.
    """

    def __init__(self, c):
        super(R2LogR2, self).__init__(c)

    def apply(self, x):
        """
        Apply the basis function.

        .. note::

            :math:`r^2 \log{r^2} === r^2 2 \log{r}`

        Parameters
        ----------
        x : (N, D) ndarray
            Set of points to apply the basis to.

        Returns
        -------
        u : (N, L) ndarray
            The basis function applied to each distance,
            :math:`\lVert x - c \rVert`.
        """
        euclidean_distance = cdist(x, self.c)
        mask = euclidean_distance == 0
        with np.errstate(divide='ignore', invalid='ignore'):
            u = (euclidean_distance ** 2 *
                 (2 * np.log(euclidean_distance)))
            # reset singularities to 0
        u[mask] = 0
        return u

    def jacobian_points(self, x):
        """
        Apply the derivative of the basis function wrt the coordinate system.
        This is applied over each dimension of the input vector, `x`.

        .. note::

            ..math::

                2 (x - c)^T (\log{r^2_{x, l}} + 1) ===
                2 (x - c)^T (2 \log{r_{x, l}} + 1)

            where:

            :math:`r_{x, l} = \lVert x - c \rVert``


        Parameters
        ----------
        x : (N, D) ndarray
            Set of points to apply the basis to.

        Returns
        -------
        dudx : (N, L, D) ndarray
            The jacobian tensor representing the first order partial derivative
            of each point wrt the coordinate system
        """
        euclidean_distance = cdist(x, self.c)
        component_distances = x[..., None, ...] - self.c
        # Avoid log(0) and set to 1 so that log(1) = 0
        euclidean_distance[euclidean_distance == 0] = 1
        dudx = (2 * component_distances *
                (2 * np.log(euclidean_distance[..., None]) + 1))
        return dudx


class R2LogR(BasisFunction):
    r"""
    Calculates the :math:`r^2 \log{r}` basis function.

    The derivative of this function is :math:`r (1 + 2 \log{r})`.

    .. note::

        :math:`r = \lVert x - c \rVert`

    Parameters
    ----------
    c : (L, D) ndarray
        The set of centers that make the basis. Usually represents a set of
        source landmarks.
    """

    def __init__(self, c):
        super(R2LogR, self).__init__(c)

    def apply(self, x):
        """
        Apply the basis function :math:`r^2 \log{r}`.

        Parameters
        ----------
        x : (N, D) ndarray
            Set of points to apply the basis to.

        Returns
        -------
        u : (N, L) ndarray
            The basis function applied to each distance,
            :math:`\lVert x - c \rVert`.
        """
        euclidean_distance = cdist(x, self.c)
        mask = euclidean_distance == 0
        with np.errstate(divide='ignore', invalid='ignore'):
            u = euclidean_distance ** 2 * np.log(euclidean_distance)
        # reset singularities to 0
        u[mask] = 0
        return u

    def jacobian_points(self, x):
        """
        The derivative of the basis function wrt the coordinate system
        evaluated at `x`.

        :math:`(x - c)^T (1 + 2 \log{r_{x, l}})`.

        .. note::

            :math:`r_{x, l} = \lVert x - c \rVert``

        Parameters
        ----------
        x : (N, D) ndarray
            Set of points to apply the basis to.

        Returns
        -------
        dudx : (N, L, D) ndarray
            The jacobian tensor representing the first order partial derivative
            of each points wrt the coordinate system
        """
        euclidean_distance = cdist(x, self.c)
        component_distances = x[..., None, ...] - self.c
        # Avoid log(0) and set to 1 so that log(1) = 0
        euclidean_distance[euclidean_distance == 0] = 1
        dudx = (component_distances *
                (1 + 2 * np.log(euclidean_distance[..., None])))
        return dudx



import abc
from copy import deepcopy
import numpy as np

from menpo.image.masked import MaskedImage
from menpo.visualize.base import ImageViewer, DepthImageHeightViewer


class AbstractSpatialImage(MaskedImage):
    r"""
    A 2D image that represents spatial data in some fashion in it's channel
    data. As a result, it contains a :class:`menpo.shape.mesh.base.TriMesh`,
    or, if a texture is provided, a
    :class:`menpo.shape.mesh.base.TexturedTriMesh`.

    Parameters
    -----------
    image_data: (M, N, ..., C) ndarray
        Array representing the spatial image pixels, with the last axis being
        channels.
    mask: (M, N, ..., L) boolean ndarray or :class:`BooleanImage`, optional
        A suitable mask for the spatial data

        Default: All true mask
    trilist: (n_tris, 3), ndarray, optional
        Triangle list for the trimesh. If None, the trilist is generation
        from all True points using Delaunay triangulation.

        Default: None
    tcoords: (n_true, 2), ndarray, optional
        Texture coordinates relating each True value of the mask to the
        texture space

        Default: If texture is provided, tcoords are generated on the
        assumption that the texture and the spatial data are in
        correspondence. If no texture, None.
    texture: :class:`Abstract2DImage` instance, optional
        A texture to be associated with the spatial data

        Default: None (no texture)
    """

    def __init__(self, image_data, mask=None, trilist=None,
                 tcoords=None, texture=None):
        super(AbstractSpatialImage, self).__init__(image_data, mask=mask)
        if self.n_dims != 2:
            raise ValueError("Trying to build an AbstractSpatialImage with {} "
                             "dimensions - has to be 2 dimensional"
            .format(self.n_dims))
        self._trilist = None
        self._tcoords = None
        self._texture = None
        self._mesh = None

        if trilist is not None:
            self._trilist = np.array(trilist, copy=True, order='C')
        if tcoords is not None:
            self._tcoords = deepcopy(tcoords)
        if texture is not None:
            self._texture = deepcopy(texture)

    @property
    def mesh(self):
        if self._mesh is None:
            self._mesh = self._create_mesh_from_shape()
        return self._mesh

    def rebuild_mesh(self):
        self._mesh = None

    def from_vector_inplace(self, vector):
        r"""
        Takes a flattened vector and updates this SpatialImage.

        Done by reshaping the vector to the correct pixels and channels.
        Note that the only region of the image that will be filled is the
        masked region.

        Calling this method will cause a cache invalidation on this spatial
        images' mesh.

        Parameters
        ----------
        vector : (``n_pixels``,)
            A flattened vector of all pixels and channels of an image.
        """
        MaskedImage.from_vector_inplace(self, vector)
        self.rebuild_mesh()

    @abc.abstractmethod
    def _generate_points(self):
        pass

    def _create_mesh_from_shape(self):
        r"""
        Creates a mesh from the spatial information.

        Parameters
        ----------
        trilist: (n_tris, 3), ndarray, optional
            Triangle list for the trimesh. If None, the trilist is generation
            from all True points using Delaunay triangulation.

            Default: None

        tcoords: (n_true, 2), ndarray, optional
            Texture coordinates relating each True value of the mask to the
            texture space

            Default: If texture is provided, tcoords are generated on the
            assumption that the texture and the spatial data are in
            correspondence. If no texture, None.

        texture: :class:`Abstract2DImage` instance, optional
            A texture to be associated with the spatial data

            Default: None (no texture)
        """
        from menpo.shape.mesh import TriMesh, TexturedTriMesh
        from scipy.spatial import Delaunay

        points = self._generate_points()

        trilist = self._trilist
        if trilist is None:
            # Delaunay the 2D surface.
            trilist = Delaunay(points[..., :2]).simplices

        if self._texture is None:
            return TriMesh(points, trilist)
        else:
            tcoords = self._tcoords
            if tcoords is None:
                tcoords = self.mask.true_indices.astype(np.float64)
                # scale to [0, 1]
                tcoords = tcoords / np.array(self.shape)
                # (s,t) = (y,x)
                tcoords = np.fliplr(tcoords)
                # move origin to top left
                tcoords[:, 1] = 1.0 - tcoords[:, 1]
            return TexturedTriMesh(points, trilist, tcoords, self._texture)

    def _view(self, figure_id=None, new_figure=False, mode='image',
              channels=None, masked=True, **kwargs):
        r"""
        View the image using the default image viewer. Before the image is
        rendered the depth values are normalised between 0 and 1. The range
        is then shifted so that the viewable range provides a reasonable
        contrast.

        Parameters
        ----------
        mode : {'image', 'mesh', 'height'}
            The manner in which to render the depth map.

            ========== =========================
            key        description
            ========== =========================
            image      View as a greyscale image
            mesh       View as a triangulated mesh
            height     View as a height map
            ========== =========================

            Default: 'image'

        Returns
        -------
        image_viewer : :class:`menpo.visualize.viewimage.ViewerImage`
            The viewer the image is being shown within
        """
        import scipy.stats

        pixels = self.pixels.copy()
        pixels[np.isinf(pixels)] = np.nan
        pixels = np.abs(pixels)
        pixels -= scipy.stats.nanmean(pixels)
        pixels /= np.nanmax(pixels)

        mask = None
        if masked:
            mask = self.mask.mask

        if mode is 'image':
            return ImageViewer(figure_id, new_figure,
                               self.n_dims, pixels,
                               channels=channels, mask=mask).render(**kwargs)
        if mode is 'mesh':
            return self.mesh._view(figure_id=figure_id, new_figure=new_figure,
                                   **kwargs)
        else:
            return self._view_extra(figure_id, new_figure, mode, mask,
                                    **kwargs)

    def _view_extra(self, figure_id, new_figure, mode, mask, **kwargs):
        if mode is 'height':
            return DepthImageHeightViewer(
                figure_id, new_figure,
                self.pixels[:, :, 2], mask=mask).render(**kwargs)
        else:
            raise ValueError("Supported mode values are: 'image', 'mesh'"
                             " and 'height'")


class ShapeImage(AbstractSpatialImage):
    r"""
    An image the represents a shape image. Due to the fact a shape image has
    an implicit spatial meaning, it also contains a
    :class:'menpo.shape.mesh.base.TriMesh`. This allows the shape image to be
    treated as an image, but expose an object that represents the shape
    as a mesh.

    Has to be a 2D image, and has to have exactly 3 channels for (X,Y,
    Z) spatial values.

    Parameters
    -----------
    image_data: (M, N, 3) ndarray
        Array representing the spatial image pixels, with the last axis being
        the spatial data per pixel.
    mask: (M, N) boolean ndarray or :class:`BooleanImage`, optional
        A suitable mask for the spatial data

        Default: All true mask
    trilist: (n_tris, 3), ndarray, optional
        Triangle list for the trimesh. If None, the trilist is generation
        from all True points using Delaunay triangulation.

        Default: None
    tcoords: (n_true, 2), ndarray, optional
        Texture coordinates relating each True value of the mask to the
        texture space

        Default: If texture is provided, tcoords are generated on the
        assumption that the texture and the spatial data are in
        correspondence. If no texture, None.
    texture: :class:`Abstract2DImage` instance, optional
        A texture to be associated with the spatial data

        Default: None (no texture)
    """

    def __init__(self, image_data, mask=None, trilist=None,
                 tcoords=None, texture=None):
        super(ShapeImage, self).__init__(image_data, mask, trilist, tcoords,
                                         texture)
        if self.n_channels != 3:
            raise ValueError("Trying to build a ShapeImage with {} channels "
                             "- has to have exactly 3 (for X, Y, "
                             "Z)".format(self.n_channels))

    @classmethod
    def blank(cls, shape, fill=0, dtype=np.float, mask=None, **kwargs):
        r"""
        Returns a blank ShapeImage

        Parameters
        ----------
        shape : tuple or list
            The shape of the image

        fill : int, optional
            The value to fill all pixels with

            Default: 0
        dtype: numpy datatype, optional
            The datatype of the image.

            Default: np.float
        mask: (M, N) boolean ndarray or :class:`BooleanImage`
            An optional mask that can be applied to the image. Has to have a
             shape equal to that of the image.

             Default: all True :class:`BooleanImage`

        Returns
        -------
        blank_image : :class:`ShapeImage`
            A new shape image of the requested size.
        """
        n_channels = kwargs.get('n_channels', 3)
        if n_channels != 3:
            raise ValueError('The number of channels of a ShapeImage must be '
                             'set to 3')
        blank_image = super(AbstractSpatialImage, cls).blank(
            shape, n_channels=n_channels, fill=fill, dtype=dtype, mask=mask)
        blank_image.pixels[:, :, :2] = np.reshape(
            blank_image.mask.all_indices, (shape[0], shape[1], 2))
        return blank_image

    def as_depth_image(self):
        """
        Convert the shape image to a depth image by stripping off the z-values.
        Copies all associated data including landmarks and texture information.

        Returns
        -------
        depth_image : :class:`DepthImage`
            The depth image created by taking the z-values of this shape image.
        """
        depth_image = DepthImage(deepcopy(self.pixels[:, :, 2]),
                                 deepcopy(self.mask),
                                 trilist=deepcopy(self._trilist),
                                 tcoords=deepcopy(self._tcoords),
                                 texture=deepcopy(self._texture))
        depth_image.landmarks = self.landmarks
        return depth_image

    def _generate_points(self):
        return self.masked_pixels


class DepthImage(AbstractSpatialImage):
    r"""
    An image the represents a depth image. Due to the fact a depth image has
    an implicit spatial meaning, a DepthImage also contains a
    :class:'menpo.shape.mesh.base.TriMesh`. This allows the depth image to be
    treated as an image, but expose an object that represents the depth
    as a mesh.

    Will have exactly 1 channel. The numpy array used to build the
    DepthImage is of shape (M, N) - it does not include the channel axis.

    In images, the origin is in the top-left and the first axis represents
    the y-direction (down the image). For ``DepthImage``, when accessing the
    mesh property, the origin is interpreted as the bottom-left and the first
    axis represents the x-direction (across the image). This is to maintain
    consistency with a right-handed coordinate scheme for meshes.

    Parameters
    -----------
    image_data: (M, N) ndarray
        Array representing the spatial image pixels. There is no channel
        axis - each pixel position stores a single depth value.
    mask: (M, N) boolean ndarray or :class:`BooleanImage`, optional
        A suitable mask for the spatial data

        Default: All true mask
    trilist: (n_tris, 3), ndarray, optional
        Triangle list for the trimesh. If None, the trilist is generation
        from all True points using Delaunay triangulation.

        Default: None
    tcoords: (n_true, 2), ndarray, optional
        Texture coordinates relating each True value of the mask to the
        texture space

        Default: If texture is provided, tcoords are generated on the
        assumption that the texture and the spatial data are in
        correspondence. If no texture, None.
    texture: :class:`Abstract2DImage` instance, optional
        A texture to be associated with the spatial data

        Default: None (no texture)
    """

    def __init__(self, image_data, mask=None, trilist=None,
                 tcoords=None, texture=None):
        super(DepthImage, self).__init__(image_data[..., None], mask, trilist,
                                         tcoords, texture)
        if self.n_channels != 1:
            raise ValueError("Trying to build a DepthImage with {} channels "
                             "- has to have exactly 1 (for Z values)"
            .format(self.n_channels))

    @classmethod
    def blank(cls, shape, fill=0, dtype=np.float, mask=None, **kwargs):
        r"""
        Returns a blank DepthImage

        Parameters
        ----------
        shape : tuple or list
            The shape of the image

        fill : int, optional
            The value to fill all pixels with

            Default: 0
        dtype: numpy datatype, optional
            The datatype of the image.

            Default: np.float
        mask: (M, N) boolean ndarray or :class:`BooleanImage`
            An optional mask that can be applied to the image. Has to have a
             shape equal to that of the image.

             Default: all True :class:`BooleanImage`

        Returns
        -------
        blank_image : :class:`DepthImage`
            A new depth image of the requested size.
        """
        n_channels = kwargs.get('n_channels', 1)
        if n_channels != 1:
            raise ValueError('The number of channels of a DepthImage must be '
                             'set to 1')
        return super(DepthImage, cls).blank(
            shape, n_channels=n_channels, fill=fill, dtype=dtype, mask=mask)

    @classmethod
    def _init_with_channel(cls, image_data_with_channel, mask):
        if image_data_with_channel.ndim != 3 or \
                        image_data_with_channel.shape[-1] != 1:
            raise ValueError("DepthImage must be constructed with 3 "
                             "dimensions and 1 channel.")
        return cls(image_data_with_channel[..., 0], mask)

    def _generate_points(self):
        # We need to make the axes consistent with a right-handed coordinate
        # scheme, so we flip the x and y axis (make the first axis the x)
        points = np.hstack((self.mask.true_indices[:, [1, 0]],
                            self.masked_pixels))
        # Then we need to ensure that the origin is in the bottom left,
        # so we flip the y-axis
        points[:, 1] = self.mask.height - points[:, 1]
        return points

    def _view_extra(self, figure_id, new_figure, mode, mask, **kwargs):
        r"""
        View the image using the default image viewer. Before the image is
        rendered the depth values are normalised between 0 and 1. The range
        is then shifted so that the viewable range provides a reasonable
        contrast.

        Parameters
        ----------
        mode : {'image', 'mesh', 'height'}
            The manner in which to render the depth map.

            ========== =========================
            key        description
            ========== =========================
            image      View as a greyscale image
            mesh       View as a triangulated mesh
            height     View as a height map
            ========== =========================

            Default: 'image'

        Returns
        -------
        image_viewer : :class:`menpo.visualize.viewimage.ViewerImage`
            The viewer the image is being shown within
        """
        if mode is 'height':
            return DepthImageHeightViewer(
                figure_id, new_figure,
                self.pixels[:, :, 0], mask=mask).render(**kwargs)
        else:
            raise ValueError("Supported mode values are: 'image', 'mesh'"
                             " and 'height'")



import numpy as np
from menpo.groupalign.base import MultipleAlignment
from menpo.shape import PointCloud
from menpo.transform import AlignmentSimilarity, UniformScale, Translation


class GeneralizedProcrustesAnalysis(MultipleAlignment):
    r"""
    TODO: Write documentation
    """
    def __init__(self, sources, **kwargs):
        super(GeneralizedProcrustesAnalysis, self).__init__(sources, **kwargs)
        self.transforms = [AlignmentSimilarity(source, self.target)
                           for source in self.sources]
        self.initial_target_scale = self.target.norm()
        self.n_iterations = 1
        self.max_iterations = 100
        self.converged = self._recursive_procrustes()
        print self

    def _recursive_procrustes(self):
        """
        Recursively calculates a Procrustes alignment
        """
        if self.n_iterations > self.max_iterations:
            return False
        av_aligned_source = sum(
            t.aligned_source.points for t in self.transforms) / self.n_sources
        new_target = PointCloud(av_aligned_source)
        # rescale the new_target to be the same size as the original about
        # it's centre
        rescale = UniformScale(
            self.initial_target_scale / new_target.norm(), self.n_dims)
        centre = Translation(-new_target.centre)
        rescale_about_centre = centre.compose_before(rescale).compose_before(
            centre.pseudoinverse)
        rescale_about_centre.apply_inplace(new_target)
        # check to see if  we have converged yet
        delta_target = np.linalg.norm(self.target.points - new_target.points)
        if delta_target < 1e-6:
            return True
        else:
            self.n_iterations += 1
            for t in self.transforms:
                t.set_target(new_target)
            self.target = new_target
            return self._recursive_procrustes()

    @property
    def mean_aligned_shape(self):
        return PointCloud(np.mean([t.target.points for t in self.transforms],
                                  axis=0))

    @property
    def av_alignment_error(self):
        return sum([t.alignment_error for t in self.transforms])/self.n_sources

    def __str__(self):
        if self.converged:
            return ('Converged after %d iterations with av. error %f'
                    % (self.n_iterations, self.av_alignment_error))
        else:
            return ('Failed to converge after %d iterations with av. error '
                    '%f' % (self.n_iterations, self.av_alignment_error))



from numpy.testing import assert_approx_equal

from menpo.image import BooleanImage
import menpo.io as pio
from menpo.transform import Affine

from menpo.fit.lucaskanade.image import (ImageInverseCompositional,
                                         ImageForwardAdditive)
from menpo.fit.lucaskanade.residual import *


target_shape = (90, 90)

# Setup the static assets (the takeo image)
takeo = pio.import_builtin_asset('takeo.ppm')
image = takeo.as_greyscale()
template_mask = BooleanImage.blank(target_shape)

# Setup global conditions
initial_params = np.array([0, 0, 0, 0, 70.5, 30.5])
target_params = np.array([0, 0.2, 0.1, 0, 70, 30])


def compute_rms_point_error(test_pts, template_affine, M):
    iteration_pts = M.apply(template_affine.T)
    diff_pts = test_pts - iteration_pts
    return np.sqrt(np.mean(diff_pts ** 2))


def setup_conditions(interpolator):
    target_transform = Affine.identity(2).from_vector(target_params)
    image_warped = image.warp_to(template_mask, target_transform,
                                 interpolator=interpolator)
    return image, image_warped, initial_params


def setup_error():
    target_transform = Affine.identity(2).from_vector(target_params)
    original_box = np.array([[0,               0],
                             [target_shape[0], 0],
                             [target_shape[0], target_shape[1]],
                             [0,               target_shape[1]]]).T
    target_pts = target_transform.apply(original_box.T)

    return target_pts, original_box


def compute_fixed_error(transform):
    target_pts, original_box = setup_error()
    return compute_rms_point_error(target_pts, original_box, transform)


def residual_wrapper(residual, algorithm, interpolator, expected_error):
    image, template, initial_params = setup_conditions(interpolator)
    align_algorithm = algorithm(
        template, residual, Affine.identity(2).from_vector(
            initial_params))
    fitting = align_algorithm.fit(image, initial_params)
    transform = fitting.final_transform
    rms_error = compute_fixed_error(transform)
    assert_approx_equal(rms_error, expected_error)

###############################################################################


def test_2d_ls_ic_map_coords():
    residual_wrapper(LSIntensity(), ImageInverseCompositional,
                     'scipy',
                     1.551662325323618e-05)


def test_2d_ls_fa_map_coords():
    residual_wrapper(LSIntensity(), ImageForwardAdditive,
                     'scipy',
                     1.4743724618438824e-05)


def test_2d_ecc_ic_map_coords():
    residual_wrapper(ECC(), ImageInverseCompositional,
                     'scipy',
                     2.0802365833550096e-06)


def test_2d_ecc_fa_map_coords():
    residual_wrapper(ECC(), ImageForwardAdditive,
                     'scipy',
                     1.9477094730117106)


def test_2d_gabor_ic_map_coords():
    global target_shape
    residual_wrapper(GaborFourier(target_shape), ImageInverseCompositional,
                     'scipy',
                     8.754889609451457)


def test_2d_gabor_fa_map_coords():
    global target_shape
    residual_wrapper(GaborFourier(target_shape), ImageForwardAdditive,
                     'scipy',
                     0.03729593561127312)


def test_2d_gradientimages_ic_map_coords():
    residual_wrapper(GradientImages(), ImageInverseCompositional,
                     'scipy',
                     10.002935866056646)


def test_2d_gradientimages_fa_map_coords():
    residual_wrapper(GradientImages(), ImageForwardAdditive,
                     'scipy',
                     9.952157644001336)


def test_2d_gradientcorrelation_ic_map_coords():
    residual_wrapper(GradientCorrelation(), ImageInverseCompositional,
                     'scipy',
                     6.93587178891484)


def test_2d_gradientcorrelation_fa_map_coords():
    residual_wrapper(GradientCorrelation(), ImageForwardAdditive,
                     'scipy',
                     5.388114566437586)
